#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Source for NGSI Agents to collect from.

Sources MUST respect the following protocol :
Each Source Class is a generator hence MUST implement __iter__().
Some Sources MAY implement close() if needed to free resources.
"""

import sys
import gzip
import json
import time
import glob

from dataclasses import dataclass
from collections.abc import Iterable
from loguru import logger
from os.path import basename
from typing import List, Callable, Tuple, Any, Sequence
from more_itertools import take
from itertools import islice, chain
from zipfile import ZipFile
from io import TextIOWrapper
from pathlib import Path

from pyngsi.utils import stream_from


@dataclass(eq=True)
class Row:
    """
    A row is a data record delivered from a Source.

    A row is composed of the record (the data itself) and the provider (the name of the datasource provider).
    For example, the provider can be the full qualified named of a remote file located on a FTP Server.
    The record could be a simple string, a CSV-delimited line, a full JSON document.
    """
    provider: str = "user"
    record: Any = None


class Source(Iterable):
    """
    A Source is a pull datasource : any datasource we can iterate on.

    The library provides many sources.
    One can code its own Source just by extending Source, and providing a new Row for each iteration.
    """

    registered_extensions = {}

    def __init__(self, rows: Sequence[Row]):
        self.rows = rows

    def __iter__(self):
        yield from self.rows

    def head(self, n: int = 10) -> List[Row]:
        """return a list built from the first n elements"""
        return take(n, self)

    def first(self) -> Row:
        """return the first element"""
        row: Row = None
        try:
            row = self.head(1)[0]
        except Exception:
            pass
        return row

    def skip_header(self, lines: int = 1):
        """return a new Source with first n lines skipped, default is to skip only the first line"""
        return Source(islice(self, lines, None))

    def limit(self, n: int = 10):
        """return a new Source limited to the first n elements"""
        iterator = iter(self)
        return Source((next(iterator) for _ in range(n)))

    @classmethod
    def from_stream(cls, stream: Iterable = sys.stdin, provider: str = "user", **kwargs):
        """automatically create the Source from a stream"""
        return SourceStream(stream, **kwargs)

    @classmethod
    def from_file(cls, filename: str, provider: str = "user", **kwargs):
        from pyngsi.sources.source_json import SourceJson
        """automatically create the Source from a filename, figuring out the extension, handles text, json and gzip compression"""
        if "*" in cls.registered_extensions:
            klass, kwargs = cls.registered_extensions["*"]
            return klass(filename, **kwargs)
        ext = (''.join(Path(filename).suffixes))[1:]
        if ext in cls.registered_extensions:
            klass, kwargs = cls.registered_extensions[ext]
            return klass(filename, **kwargs)
        stream, suffixes = stream_from(filename)
        ext = suffixes[-1]
        if ext == ".json":
            json_obj = json.load(stream)
            return SourceJson(json_obj, provider=basename(filename), **kwargs)
        return SourceStream(stream, provider=basename(filename), **kwargs)

    @classmethod
    def from_files(cls, filenames: Sequence[str], provider: str = "user", **kwargs):
        sources = [Source.from_file(f) for f in filenames]
        return SourceMany(sources)

    @classmethod
    def from_glob(cls, pattern: str, provider: str = "user", **kwargs):
        sources = [Source.from_file(f) for f in glob.glob(pattern)]
        return SourceMany(sources)   

    @classmethod
    def from_globs(cls, patterns: Sequence[str], provider: str = "user", **kwargs):
        filenames = chain.from_iterable([glob.glob(p) for p in patterns])
        sources = [Source.from_file(f) for f in filenames]
        return SourceMany(sources)        

    @classmethod
    def register_extension(cls, ext: str, src, **kwargs):
        cls.registered_extensions[ext] = (src, kwargs)

    @classmethod
    def unregister_extension(cls, ext: str):
        if cls.is_registered_extension(ext):
            del cls.registered_extensions[ext]

    @classmethod
    def is_registered_extension(cls, ext: str):
        return ext in cls.registered_extensions

    @classmethod
    def register(cls, src, **kwargs):
        cls.register_extension("*", src, **kwargs)

    @classmethod
    def unregister(cls):
        cls.unregister_extension("*")

    def reset(self):
        pass


class SourceStream(Source):

    def __init__(self, stream: Iterable, provider: str = "user", ignore_header: bool = False):
        if ignore_header:
            next(stream)
        self.stream = stream
        self.provider = provider

    def __iter__(self):
        for line in self.stream:
            yield Row(self.provider, line.rstrip("\r\n"))

    def reset(self):
        pass


class SourceStdin(SourceStream):

    def __init__(self, **kwargs):
        super().__init__(stream=sys.stdin, **kwargs)


class SourceSingle(Source):

    """
    A SourceSingle is Source built from a Python string.

    """

    def __init__(self, row: Any, provider: str = "user"):
        self.row = row
        self.provider = provider

    def __iter__(self):
        yield Row(self.provider, self.row)


class SourceMany(Source):

    def __init__(self, sources: Sequence[Source], provider: str = "user"):
        self.sources = sources
        self.provider = provider

    def __iter__(self):
        for src in self.sources:
            yield from src



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys

from dataclasses import dataclass
from shortuuid import uuid
from loguru import logger
from datetime import datetime
from typing import Callable, Union
from abc import ABC, abstractmethod

from pyngsi.sources.source import Row, Source, SourceStream
from pyngsi.sink import Sink, SinkStdout
from pyngsi.ngsi import DataModel
from pyngsi.sources.server import Server
from pyngsi.__init__ import __version__


class NgsiException(Exception):
    pass


class NgsiAgent(ABC):

    """
    Abstract Class used by either NgsiAgentPull or NgsiAgentServer.
    """

    @property
    @abstractmethod
    def status(self):
        pass

    @staticmethod
    def create_agent(src: Union[Source, Server] = SourceStream(sys.stdin),
                     sink: Sink = SinkStdout(),
                     process: Callable = lambda x: x.record,
                     side_effect: Callable[[Row, Sink, DataModel], int] = None):
        """
        Factory method to create the agent depending on the source push/pull.

        :param src: the Source
        :param sink: the Sink
        :param process: a function that takes an input row from the source and outputs a NGSI datamodel
        """
        if isinstance(src, Source):
            return NgsiAgentPull(src, sink, process, side_effect)
        elif isinstance(src, Server):
            return NgsiAgentServer(src, sink, process, side_effect)
        else:
            raise NgsiException(
                f"Cannot create agent. Unknown source type {type(src)}")

    @abstractmethod
    def run(self):
        """
        Run the NGSI Agent
        """
        pass

    @abstractmethod
    def close(self):
        """
        Close the NGSI Agent
        """
        pass

    @dataclass(eq=True)
    class Stats:
        """
        Agent processing statistics
        """
        input: int = 0
        processed: int = 0
        output: int = 0
        filtered: int = 0
        error: int = 0
        side_entities: int = 0

        def __add__(self, o):
            return NgsiAgent.Stats(self.input + o.input,
                                   self.processed + o.processed,
                                   self.output + o.output,
                                   self.filtered + o.filtered,
                                   self.error + o.error,
                                   self.side_entities + o.side_entities)

        def __iadd__(self, o):
            self.input += o.input
            self.processed += o.processed
            self.output += o.output
            self.filtered += o.filtered
            self.error += o.error
            self.side_entities += o.side_entities
            return self

        def zero(self):
            self.input = 0
            self.processed = 0
            self.output = 0
            self.filtered = 0
            self.error = 0
            self.side_entities = 0
            return self

class NgsiAgentPull(NgsiAgent):

    """
    The NgsiAgentPull pulls rows from the datasource
    """

    def __init__(self,
                 source: Source = None,
                 sink: Sink = None,
                 process: Callable = lambda row, *args, **kwargs: row.record,
                 side_effect: Callable = None):
        logger.info("init NGSI agent")
        self.source = source if source else SourceStream(sys.stdin)
        logger.info(f"source = [{self.source.__class__.__name__}]")
        self.sink = sink if sink else SinkStdout()
        logger.info(f"sink = [{self.sink.__class__.__name__}]")
        self.process = process
        self.side_effect = side_effect
        self.stats = NgsiAgent.Stats()

    @property
    def status(self):
        return self.stats

    def run(self):
        logger.info("start to acquire data")
        for row in self.source:
            logger.debug(row)
            try:
                if row.provider is None:
                    row.provider = "user"
                logger.trace(f"{row.provider=}\t{row.record=}")
                self.stats.input += 1
                x = self.process(row)
                if not x:
                    self.stats.filtered += 1
                    continue
                self.stats.processed += 1
                msg = x.json() if isinstance(x, DataModel) else x
                self.sink.write(msg)
                self.stats.output += 1
                if self.side_effect:
                    side_entities = self.side_effect(row, self.sink, x)
                    self.stats.side_entities += side_entities
            except Exception as e:
                self.stats.error += 1
                logger.error(f"Cannot process record : {e}")
        return self

    def close(self):
        logger.info("close NGSI agent")
        logger.info(self.status)
        # logger.info(f"close source")
        # self.source.close()
        logger.info(f"close sink")
        self.sink.close()

    def reset(self):
        self.source.reset()
        self.stats.zero()


class NgsiAgentServer(NgsiAgent):

    """
    The NgsiAgentServer processes data from a push datasource.

    The NgsiAgentServer acts both as a Source and as an Agent.
    In this case, the Source is a Server that listens from incoming data.
    Each time the Source receives a data request, the NgsiAgentServer is triggered to process the request content.
    """

    import pyngsi.sources.server

    @dataclass
    class ServerStatus:
        version = __version__
        starttime: datetime = None
        lastcalltime: datetime = None
        calls: int = 0
        calls_success: int = 0
        calls_error: int = 0

        def __init__(self):
            self.starttime = datetime.now()

    def __init__(self,
                 server: pyngsi.sources.server.Server = None,
                 sink: Sink = None,
                 process: Callable = lambda row, *args, **kwargs: row.record,
                 side_effect: Callable[[Row, Sink, DataModel], int] = None):
        logger.info("init NGSI agent")
        self.server = server
        logger.info(f"server = [{self.server.__class__.__name__}]")
        self.sink = sink if sink else SinkStdout()
        logger.info(f"sink = [{self.sink.__class__.__name__}]")
        self.process = process
        logger.info(f"process = [{self.process}]")
        self.side_effect = side_effect
        logger.info(f"side_effect = [{self.side_effect}]")
        self.server_status = self.ServerStatus()
        self.stats = NgsiAgent.Stats()

    @property
    def status(self):
        return self.server_status, self.stats

    def run(self):
        logger.info("start server")
        self.start_time = datetime.now()

        server: Server = self.server
        server.set_agent(self)  # a NGSI server also acts as an agent
        logger.info(f"{server.agent=}")

        self.server.run()
        return self

    def close(self):
        logger.info("close NGSI agent")
        logger.info(self.status)
        logger.info(f"close server")
        self.server.close()
        logger.info(f"close sink")
        self.sink.close()


def build_entity_unknown(row: Row) -> DataModel:
    """
    Helper function to quickly build a datamodel without the need of id and type.
    """
    m = DataModel(id=f"Unknown:{uuid()}", type="Unknown")
    m.add("rawData", row.record)
    return m


def build_entity_sample_orion(row: Row) -> DataModel:
    """
    Helper function to build a datamodel for the NGSI walkthrough tutorial.

    Please have a look at : https://fiware-orion.readthedocs.io/en/master/user/walkthrough_apiv2/index.html#entity-creation
    Here we consider the input as a CSV line.
    """
    id, temperature, pressure = row.record.split(';')
    m = DataModel(id=id, type="Room")
    m.add("temperature", float(temperature))
    m.add("pressure", int(pressure))
    return m



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import socket
import signal
import time
import schedule
import _thread

from flask import Flask, request, jsonify
from cheroot.wsgi import Server as WSGIServer
from loguru import logger
from datetime import datetime
from typing import Callable
from dataclasses import dataclass
from enum import Enum, auto

from pyngsi.sink import Sink
from pyngsi.agent import NgsiAgent, NgsiAgentPull
from pyngsi.__init__ import __version__


class UNIT(Enum):
    seconds = "s"
    minutes = "m"
    hours = "h"
    days = "d"


class SchedulerException(Exception):
    pass


@dataclass
class SchedulerStatus:
    version = __version__
    starttime: datetime = None
    lastcalltime: datetime = None
    calls: int = 0
    calls_success: int = 0
    calls_error: int = 0
    stats: NgsiAgent.Stats = None

    def __init__(self):
        self.starttime = datetime.now()
        self.stats = NgsiAgent.Stats()


class Scheduler():
    """
    Poll takes an agent and polls at periodic intervals.

    Poll updates statitics and provides information (status and version) 
    """
    def __init__(self,
                 agent: NgsiAgentPull,
                 host: str = "0.0.0.0",
                 port: int = 8081,
                 wsgi_port: int = 8880,
                 debug: bool = False,
                 interval: int = 1,
                 unit: UNIT = UNIT.minutes):

        self.agent = agent
        self.host = host
        self.port = port
        self.wsgi_port = wsgi_port
        self.debug = debug
        self.interval = interval
        self.unit = unit
        self.status = SchedulerStatus()

        self.app = Flask(__name__)
        self.app.add_url_rule("/version", 'version',
                              self._version, methods=['GET'])
        self.app.add_url_rule("/status", 'status',
                              self._status, methods=['GET'])

    def _flaskthread(self):
        if self.debug:
            self.app.run(host=self.host, port=self.port, debug=self.debug)
        else:
            wsgi_server = WSGIServer(bind_addr=(
                "0.0.0.0", self.wsgi_port), wsgi_app=self.app, numthreads=100)
            try:
                wsgi_server.start()
            except KeyboardInterrupt:
                wsgi_server.stop()

    def _job(self):
        logger.info(f"start new job at {datetime.now()}")
        self.status.lastcalltime = datetime.now()
        self.status.calls += 1

        # run the NGSI Agent
        try:
            self.agent.run()
        except Exception as e:
            logger.error(f"Error while running job : {e}")
            self.status.calls_error += 1
        else:
            self.status.calls_success += 1

        logger.info(self.agent.stats)

        self.status.stats += self.agent.stats
        self.agent.reset()

    def run(self):
        logger.info(
            f"HTTP server listens on http://{self.host}:{self.port}")
        self.status.starttime = datetime.now()
        _thread.start_new_thread(self._flaskthread, ())

        logger.info("run job now")
        self._job()
        logger.info("schedule job")
        if self.unit == UNIT.seconds:
            schedule.every(self.interval).seconds.do(self._job)
            tick = 1
        elif self.unit == UNIT.minutes:
            schedule.every(self.interval).minutes.do(self._job)
            tick = 4
        elif self.unit == UNIT.hours:
            schedule.every(self.interval).hours.do(self._job)
            tick = 32
        elif self.unit == UNIT.days:
            schedule.every(self.interval).days.do(self._job)
            tick = 128

        while True:
            logger.trace("tick")
            schedule.run_pending()
            time.sleep(tick)

    def _version(self):
        logger.trace("ask for version")
        return jsonify(name="pyngsi", version=__version__)

    def _status(self):
        logger.trace("ask for status")
        remote_status = self.agent.sink.status()
        if remote_status:
            return jsonify(poll_status=self.status,
                           orion_status=remote_status)
        else:
            return jsonify(poll_status=self.status)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Sinks.

Sinks MUST respect the following protocol :
Each Sink Class MUST implement write().
Some Sinks MAY override close() if needed to free resources.

SinkOrion is the one you will want to use in your project.
Other sinks such as SinkStdout or SinkFile are useful during the development stage and for unit testing.
"""


import gzip
import requests
import os
import copy

from abc import ABC, abstractmethod
from loguru import logger
from requests_toolbelt.utils import dump

from pyngsi.__init__ import __version__ as version


class Sink(ABC):
    """
    Sink is an abstract class

    The library provides many sinks.
    One can code its own Sink just by extending Sink.
    """

    @abstractmethod
    def write(self, msg):
        pass

    def status(self):
        pass

    def close(self):
        pass


class SinkException(Exception):
    pass


class SinkNull(Sink):
    """Do not write anything. For debugging purpose only."""

    def write(self, msg):
        pass


class SinkStdout(Sink):
    """Write to Standard Output"""

    def write(self, msg):
        print(msg)

# TODO : add append option


class SinkFile(Sink):
    """Write to file"""

    def __init__(self, filename, append=False):
        """
        Parameters
        ----------
        filename : str
            The name of the output file
        """
        self.filename = filename
        try:
            self.file = open(
                self.filename, "a" if append else "w", encoding="utf8")
        except Exception as e:
            raise SinkException(f"cannot open file {self.filename} : {e}")

    def write(self, msg: str):
        try:
            self.file.write(f"{msg}{os.linesep}")
        except Exception as e:
            raise SinkException(f"cannot write to file {self.filename} : {e}")

    def close(self):
        try:
            self.file.close()
        except Exception as e:
            raise SinkException(f"cannot close file {self.filename} : {e}")


class SinkFileGzipped(SinkFile):
    """Write to gzipped file"""

    def __init__(self, filename):
        """
        Parameters
        ----------
        filename : str
            The name of the output file
        """
        self.filename = filename
        try:
            self.file = gzip.open(self.filename, "wt", encoding="utf8")
        except Exception as e:
            raise SinkException(f"cannot open file {self.filename} : {e}")


class SinkHttp(Sink):
    """Send to HTTP server

    Attributes
    ----------
    hostname : str
        Server hostname
    port : int
        Server port
    baseurl: str
        Server Base URL
    useragent: str
        HTTP User-Agent header sent in the request
    status_endpoint: str
        endpoint to ask server for its status and its processing statistics        
    proxy: str
        HTTP Proxy string (i.e http://127.0.0.1:8080)
    """

    def __init__(self, hostname="127.0.0.1", port=8080, secure=False, baseurl="/",
                 post_endpoint="/", post_query="", status_endpoint="/status",
                 useragent=f"NgsiAgent v{version}",
                 proxy=None):
        """
        Parameters
        ----------
        hostname : str
            Server hostname
        port : int
            Server port
        baseurl: str
            Server Base URL
        useragent: str
            HTTP User-Agent header sent in the request
        proxy: str
            HTTP Proxy string (i.e http://127.0.0.1:8080)
        """
        logger.debug("init SinkHttp")
        if (baseurl[0] != "/"):
            raise Exception("baseurl must begin with a slash")

        self.hostname = hostname
        self.port = port
        self.protocol = "https" if secure else "http"
        self.baseurl = baseurl = baseurl.rstrip("/")
        self.post_endpoint = post_endpoint = post_endpoint.rstrip("/")
        self.status_endpoint = status_endpoint = status_endpoint.rstrip("/")
        prefix = f"{self.protocol}://{hostname}:{port}{baseurl}"
        self.post_url = f"{prefix}{post_endpoint}?{post_query}" if post_query else f"{prefix}{post_endpoint}"
        self.status_url = f"{prefix}{status_endpoint}"
        self.proxy = proxy
        self.headers = {'Content-Type': 'application/json',
                        'User-Agent': useragent}
        logger.info(f"{self.baseurl=}")
        logger.info(f"{self.post_url=}")
        logger.info(f"{self.status_url=}")
        logger.info(f"{useragent=}")
        logger.info(f"{self.proxy=}")

    def write(self, msg):
        """Sends HTTP POST request with the NGSI data

        Parameters
        ----------
        msg: str
            the NGSI data
        """

        try:
            r = requests.post(
                self.post_url, msg, headers=self.headers,
                proxies={self.proxy} if self.proxy else None)
            logger.trace(dump.dump_all(r).decode('utf-8'))
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise SinkException(
                f"cannot write to SinkHttp : {e}\nServer returned : {r.text}\nrecord={msg}")
        except Exception as e:
            raise SinkException(
                f"cannot write to SinkHttp : {e}\nrecord={msg}")

    def status(self) -> dict:
        logger.debug("ask http server status")
        try:
            if 'Content-Type' in self.headers: # workaround unwanted Content-Type
                headers = self.headers.copy()
                del headers['Content-Type']
            else:
                headers = self.headers()
            r = requests.get(self.status_url, headers=headers)
            logger.trace(dump.dump_all(r).decode('utf-8'))
            r.raise_for_status()
            return r.json()
        except requests.exceptions.HTTPError as e:
            logger.error(e)
            orion_status = {}
            orion_status['state'] = 'Down or Unreachable'
            orion_status['exception'] = e
            orion_status['server_message'] = r.text
            return orion_status
        except Exception as e:
            logger.error(e)
            orion_status = {}
            orion_status['state'] = 'Down or Unreachable'
            return orion_status


class SinkOrion(SinkHttp):
    """Send to Orion Context Broker"""

    def __init__(self, hostname="127.0.0.1", port="1026", secure=False, baseurl="/",
                 post_endpoint="/v2/entities", post_query="options=upsert", status_endpoint="/version",
                 useragent=f"NgsiAgent v{version}", proxy=None,
                 token=None, service=None, servicepath=None):
        logger.debug("init SinkOrion")
        super().__init__(hostname, port, secure, baseurl,
                         post_endpoint, post_query, status_endpoint,
                         useragent, proxy)
        if 'X-Auth-Token' in self.headers:
            logger.info(
                "A token has already been provided to the pyngsi framework.")
        else:
            if token:
                logger.info("A token has been set programmatically.")
            elif token := os.environ.get('ORION_TOKEN', None):
                logger.info("A token has been found in environment variable.")
            else:
                try:
                    with open('/run/secrets/orion_token') as f:
                        token = f.read()  # 2nd try to get the token from a given file (Docker Secret mode)
                        token = token.rstrip(" \r\n")
                except Exception:
                    pass
                else:
                    logger.info("A token has been found in docker secrets.")
            if token:  # a token has been found
                logger.info("Use token for authentication")
                # add the token to the Authorization header
                self.headers['X-Auth-Token'] = token
            else:
                logger.info(
                    "No token found. Request Orion without authentication.")

        if service is not None:
            self.headers['Fiware-Service'] = service
        if servicepath is not None:
            self.headers['Fiware-ServicePath'] = servicepath



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import json
import gzip

from typing import Tuple, List, Callable
from loguru import logger
from os.path import basename
from zipfile import ZipFile
from io import TextIOWrapper

from pyngsi.sources.source import Row, Source


class SourceJson(Source):
    """Read JSON formatted data from Standard Input"""

    def __init__(self, input: str, provider: str = "user", jsonpath: str = None):
        self.json_obj = input
        self.provider = provider
        self.path = jsonpath

    def __iter__(self):
        obj = self.json_obj
        if self.path:
            obj = self.jsonpath(self.path)

        if isinstance(obj, list):
            for j in obj:
                yield Row(self.provider, j)
        else:
            yield Row(self.provider, obj)

    def jsonpath(self, path: List):
        obj = self.json_obj
        for p in path:
            if isinstance(p, int):
                obj = obj[p]
            else:
                obj = obj.get(p)
        return obj

    def reset(self):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import Tuple, List, Callable
from loguru import logger
from pyngsi.ftpclient import FtpClient

from pyngsi.sources.source import Source


# a file downloaded from FTP : (local_filename, remote_filename)
FtpFile = Tuple[str, str]


class SourceFtp(Source):
    """
    A SourceFtp reads data from a given FTP Server.

    All the complexity is hidden for the end user.
    The SourceFtp can automatically download the desired files from the FTP Server.
    Selection of the remote files is based on the filenames, and operates inside one or many remote folders.
    The selection is operated thanks to the f_match() function that could be a regex or whatever you want.
    Once the files are downloaded (into a temp dir), the connection to the FTP Server is closed.
    Then the Source reads the downloaded files to deliver rows as usual, by iterating on file records.
    At the end, when the Source is closed, the temp dir is cleaned.
    """

    def __init__(self, host: str, user: str = "anonymous",
                 passwd: str = "guest",
                 paths: List[str] = ["/pub"],
                 use_tls: bool = False,
                 f_match: Callable[[str], bool] = lambda x: False,
                 provider: str = "user",
                 source_factory=Source.from_file):
        """
        Parameters
        ----------
        filename : str
            The name of the file containing raw data
        """

        self.host = host
        self.user = user
        self.passwd = passwd
        self.use_tls = use_tls
        self.paths = paths
        self.f_match = f_match
        self.provider = provider
        self.source_factory = source_factory

        # connect to FTP server
        self.ftp = FtpClient(host, user, passwd, use_tls)

        # retrieve a list of files we're interested in
        remote_files = self._retrieve_filelist(paths, f_match)

        # download files : a list of (local_filename, remote_filename)
        self.downloaded_files: List[FtpFile] = self._download_files(
            remote_files)

        if len(self.downloaded_files) != len(remote_files):
            logger.critical(f"Some files have not been downloaded.")

        # disconnect from FTP server
        self.ftp.close()

    def __iter__(self):
        for ftpfile in self.downloaded_files:
            localname, remotename = ftpfile
            logger.info(f"process local {localname}")
            provider = self.provider if self.provider else f"ftp://{self.host}{remotename}"
            source = self.source_factory(localname, provider)
            yield from source
        self.ftp.clean()

    def _retrieve_filelist(self, paths, f_match=lambda x: True) -> List[str]:
        remote_files = []
        for path in paths:
            filelist = [x for x in self.ftp.retrieve_filelist(
                path) if f_match(x)]
            remote_files.extend(filelist)
        logger.info(f"Found {len(remote_files)} matching files")
        return remote_files

    def _download_files(self, remote_files: List[str]) -> List[FtpFile]:
        try:
            downloaded_files = [(self.ftp.download(remote), remote)
                                for remote in remote_files]
        except Exception as e:
            logger.critical(f"Problem while downloading files : {e}")
        return downloaded_files

    def reset(self):
        self.__init__(self.host, self.user, self.passwd,
                      self.paths, self.f_match, self.provider, self.source_factory)



import sys
import time
import random
import openpyxl
import csv


from pathlib import Path
from loguru import logger

from pyngsi.sources.source import Source, Row


class SourceSampleOrion(Source):

    """
    A SourceSampleOrion implements the Source from the NGSI Walkthrough tutorial.

    Please have a look at :
    https://fiware-orion.readthedocs.io/en/master/user/walkthrough_apiv2/index.html#entity-creationhttps://fiware-orion.readthedocs.io/en/master/user/walkthrough_apiv2/index.html#entity-creation

    First two records are those of the tutorial.
    Following records are randomized.
    """

    def __init__(self, count: int = 5, delay: float = 1.0):
        self.count = count if count > 0 else sys.maxsize
        self.delay = delay

    def __iter__(self):
        i: int = 0
        if self.count >= 1:  # 1st element is fixed
            yield Row("orionSample", "Room1;23;720")
            i += 1
            time.sleep(self.delay)
        if self.count >= 2:  # 2nd element is fixed
            yield Row("orionSample", "Room2;21;711")
            i += 1
            time.sleep(self.delay)
        # next elements are randomized
        while i < self.count:
            yield Row("orionSample", f"Room{i%9+1};{round(random.uniform(-10,50), 1)};{random.randint(700,1000)}")
            i += 1
            time.sleep(self.delay)

    def reset(self):
        pass


class SourceMicrosoftExcel(Source):

    def __init__(self, filename, sheetid: int = 0, sheetname: str = None, ignore: int = 0):
        logger.debug(f"{filename=}")
        wb = openpyxl.load_workbook(filename, data_only=True)
        ws = wb[sheetname] if sheetname else wb.worksheets[sheetid]
        self.rows = ws.rows
        self.provider = Path(filename).name
        for _ in range(ignore):  # skip lines
            next(self.rows)

    def __iter__(self):
        for row in self.rows:
            record = ";".join(
                [str(cell.value) if cell.value else "" for cell in row])
            logger.debug(f"{self.provider=}{record=}")
            yield Row(self.provider, record)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from pyngsi.agent import NgsiAgentPull


def main():
    agent = NgsiAgentPull()
    agent.run()
    agent.close()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import ftplib
import ssl
import tempfile
import shutil

from loguru import logger
from ftplib import FTP, FTP_TLS
from typing import List
from os.path import basename, join

# https://stackoverflow.com/questions/14659154/ftpes-session-reuse-required
class MyFTP_TLS(ftplib.FTP_TLS):
    """Explicit FTPS, with shared TLS session"""

    def ntransfercmd(self, cmd, rest=None):
        conn, size = ftplib.FTP.ntransfercmd(self, cmd, rest)
        if self._prot_p:
            conn = self.context.wrap_socket(conn,
                                            server_hostname=self.host,
                                            session=self.sock.session)  # this is the fix
        return conn, size


class FtpClientException(Exception):
    pass


class FtpClient():

    def __init__(self, host: str, user: str = "anonymous",
                 passwd: str = "guest", use_tls: bool = False):
        logger.debug("Connect to FTP server")
        if use_tls:
            self.ftp = MyFTP_TLS(host)
            self.ftp.ssl_version = ssl.PROTOCOL_TLS
        else:
            self.ftp = FTP(host)
        try:
            self.ftp.login(user, passwd)
            self.ftp.set_pasv(True)
            if use_tls:
                self.ftp.prot_p()
        except ftplib.error_perm as e:
            error_code = int(str(e).split()[0])
            if (error_code == 503):
                logger.info("Already connected. Continue.")
            else:
                raise FtpClientException(f"Cannot connect : {e}")
        except ftplib.all_errors as e:
            raise FtpClientException(f"Cannot connect : {e}")
        try:
            # create temp dir to receive future downloads
            self.tmpdir = tempfile.mkdtemp()
        except Exception as e:
            logger.critical(f"Cannot create temp dir : {e}")

    def retrieve_filelist(self, path: str) -> List[str]:
        filelist: List[str] = []
        self.ftp.retrlines(f"NLST {path}", filelist.append)
        return filelist

    def download(self, remote: str) -> str:
        logger.debug(f"Download file {remote}")
        try:
            local = join(self.tmpdir, basename(remote))
            handle = open(local, 'wb')
            self.ftp.retrbinary(f"RETR {remote}", handle.write)
            handle.close()
        except Exception as e:
            raise FtpClientException(f"Cannot download {remote} : {e}")
        return local

    def close(self):
        logger.debug("Disconnect from FTP server")
        try:
            self.ftp.quit()
            self.ftp.close()
        except Exception as e:
            raise FtpClientException(f"Cannot disconnect : {e}")

    def clean(self):
        if self.tmpdir:
            try:
                shutil.rmtree(self.tmpdir)
            except Exception as e:
                logger.error(f"Cannot remove temp directory : {e}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import gzip

from zipfile import ZipFile
from io import TextIOWrapper
from pathlib import Path
from loguru import logger


def stream_from(filename: str = None):
    try:
        suffixes = Path(filename).suffixes
        ext = suffixes[-1]
        if ext == ".gz":
            stream = gzip.open(filename, "rt", encoding="utf-8")
            return stream, suffixes[:-1]
        elif ext == ".zip":
            zf = ZipFile(filename, 'r')
            f = zf.namelist()[0]
            stream = TextIOWrapper(zf.open(f, 'r'), encoding='utf-8')
            return stream, suffixes[:-1]
        else:
            return open(filename, "r", encoding="utf-8"), suffixes
    except Exception as e:
        logger.error(f"Cannot open file {filename} : {e}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyngsi",
    version="2.1.7",
    description="NGSI Python framework intended to build a Fiware NGSI Agent",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pixel-ports/pyngsi",
    author="Fabien Battello",
    author_email="fabien.battello@orange.com",
    license="Apache 2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9"
    ],
    packages=setuptools.find_packages(),
    include_package_data=False,
    install_requires=["loguru", "requests", "requests-toolbelt", "shortuuid",
                      "more_itertools", "geojson", "flask", "cherrypy", "schedule", "openpyxl"],
    test_requires=["pytest", "pytest-mock", "requests-mock", "pytest-flask"],
    python_requires=">=3.8"
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import urllib.parse

from datetime import datetime, timedelta
from geojson import Point
from typing import Any
from collections.abc import Sequence, Callable

ONE_WEEK = 7*86400


class NgsiException(Exception):
    pass


def escape(value: str) -> str:
    return urllib.parse.quote(value)


def unescape(value: str) -> str:
    return urllib.parse.unquote(value)


class DataModel(dict):

    transient_timeout = None

    def __init__(self, id: str, type: str, serializer: Callable = str):
        self.serializer = serializer
        self["id"] = id
        self["type"] = type
        if self.transient_timeout:
            self.add_transient(self.transient_timeout)

    @classmethod
    def set_transient(cls, timeout: int = ONE_WEEK):
        cls.transient_timeout = timeout

    @classmethod
    def unset_transient(cls):
        cls.transient_timeout = None

    def add(self, name: str, value: Any,
            isdate: bool = False, isurl: bool = False, urlencode=False, metadata: dict = {}):
        if isinstance(value, str):
            if isdate:
                t = "DateTime"
            elif isurl:
                t = "URL"
            elif urlencode:
                t = "STRING_URL_ENCODED"
            else:
                t = "Text"
            v = escape(value) if urlencode else value
        elif isinstance(value, bool):
            t, v = "Boolean", value
        elif isinstance(value, int):
            t, v = "Integer", value
        elif isinstance(value, float):
            t, v = "Float", value
        elif isinstance(value, datetime):
            # the value datetime MUST be UTC
            t, v = "DateTime", value.strftime("%Y-%m-%dT%H:%M:%SZ")
        elif isinstance(value, Point):
            t, v = "geo:json", value
        elif isinstance(value, tuple) and len(value) == 2:
            lat, lon = value
            try:
                location = Point((lon, lat))
            except Exception as e:
                raise NgsiException(f"Cannot create geojson field : {e}")
            t, v = "geo:json", location
        elif isinstance(value, Sequence):
            t, v = "Array", value
        elif isinstance(value, dict):
            t, v = "Property", value
        else:
            raise NgsiException(
                f"Cannot map {type(value)} to NGSI type. {name=} {value=}")
        self[name] = {"value": v, "type": t}
        if metadata:
            self[name]["metadata"] = metadata

    def add_date(self, *args, **kwargs):
        self.add(isdate=True, *args, **kwargs)

    def add_now(self, *args, **kwargs):
        self.add(value=datetime.utcnow(), *args, **kwargs)

    def add_url(self, *args, **kwargs):
        self.add(isurl=True, *args, **kwargs)

    def add_relationship(self, rel_name: str, ref_type: str, ref_id: str):
        if not rel_name.startswith("ref"):
            raise NgsiException(
                f"Bad relationship name : {rel_name}. Relationship attributes must use prefix 'ref'")
        t, v = "Relationship", f"urn:ngsi-ld:{ref_type}:{ref_id}"
        self[rel_name] = {"value": v, "type": t}

    def add_address(self, value: dict):
        t, v = "PostalAddress", value
        self["address"] = {"value": v, "type": t}

    def add_transient(self, timeout: int = ONE_WEEK, expire: datetime = None):
        if not expire:
            expire = datetime.utcnow() + timedelta(seconds=timeout)
        self.add("dateExpires", expire)

    def json(self):
        """Returns the datamodel in json format"""
        return json.dumps(self, default=self.serializer, ensure_ascii=False)

    def pprint(self):
        """Returns the datamodel pretty-json-formatted"""
        print(json.dumps(self, default=self.serializer, indent=2))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import json
import socket
import signal
import time


from flask import Flask, request, jsonify
from cheroot.wsgi import Server as WSGIServer
from loguru import logger
from datetime import datetime
from pathlib import Path
from werkzeug.utils import secure_filename

from pyngsi.sources.source import Source, SourceStream, SourceSingle
from pyngsi.sources.source_json import SourceJson

from pyngsi.__init__ import __version__ as version


class ServerException(Exception):
    pass


class Server():
    """
    A Server acts both as a Source and as an Agent

    In fact, each time the server receives a request, the server endorses the role of an agent.
    The request content is a source, the server spawns an agent to process it.
    The Server computes is own stats (i.e. incoming requests) and compute ngsi-stats related to the data processing.
    Stats are available at the end when calling the close() method, or at any time, when calling the endpoint status if available.
    """

    def __init__(self, provider="user", ignore_header=False, jsonpath=None):
        self.agent = None
        self.provider = provider
        self.ignore_header = ignore_header
        self.jsonpath = jsonpath

    def set_agent(self, agent):
        self.agent = agent

    def status(self):
        pass

    def close(self):
        pass

    def _process_content(self, src: Source):
        logger.info(f"{src=}")
        from pyngsi.agent import NgsiAgentPull
        if not src:
            logger.info("no source")
            return
        if not self.agent:
            logger.info("agent not set")
            return
        try:
            if self.ignore_header:
                src = src.skip_header()
            agent = NgsiAgentPull(src, self.agent.sink,
                                  self.agent.process, self.agent.side_effect)
            logger.info(f"{self.ignore_header=}")
            logger.info(f"{self.jsonpath=}")
            agent.run()
            agent.close()
            if self.agent:
                self.agent.stats += agent.stats
            return agent.stats
        except Exception as e:
            logger.error(f"cannot parse content : {e}")
            raise ServerException("cannot parse content")


class ServerHttpUpload(Server):
    """
    ServerHttpUpload allows receiving data from HTTP clients

    ServerHttpUpload handles raw binary (curl --data) and multipart/form-data (curl --form).
    ServerHttpUpload handles formats text and json.
    """

    def __init__(self,
                 host: str = "0.0.0.0",
                 port: int = 8081,
                 wsgi_port: int = 8880,
                 endpoint: str = "/upload",
                 debug: bool = False,
                 provider: str = None,
                 ignore_header: bool = False,
                 jsonpath: str = None):

        super().__init__(provider, ignore_header, jsonpath)
        self.host = host
        self.port = port
        self.wsgi_port = wsgi_port
        self.endpoint = endpoint
        self.debug = debug

        self.app = Flask(__name__)
        self.app.add_url_rule("/version", 'version',
                              self._version, methods=['GET'])
        self.app.add_url_rule("/status", 'status',
                              self._status, methods=['GET'])
        self.app.add_url_rule(endpoint, 'upload',
                              self._upload, methods=['POST'])

    def run(self):
        logger.info(
            f"HTTP server listens on http://{self.host}:{self.port}{self.endpoint}")
        if self.agent:
            self.agent.server_status.starttime = datetime.now()

        if self.debug:
            self.app.run(host=self.host, port=self.port, debug=self.debug)
        else:
            wsgi_server = WSGIServer(bind_addr=(
                "0.0.0.0", self.wsgi_port), wsgi_app=self.app, numthreads=100)
            try:
                wsgi_server.start()
            except KeyboardInterrupt:
                wsgi_server.stop()

    def _version(self):
        logger.trace("ask for version")
        return jsonify(name="pyngsi", version=version)

    def _status(self):
        logger.trace("ask for status")
        remote_status = self.agent.sink.status()
        if remote_status:
            return jsonify(server_status=self.agent.server_status,
                           ngsi_stats=self.agent.stats,
                           orion_status=remote_status)
        else:
            return jsonify(server_status=self.agent.server_status,
                           ngsi_stats=self.agent.stats)

    def _upload(self):
        
        if self.agent:
            self.agent.server_status.lastcalltime = datetime.now()
            self.agent.server_status.calls += 1

        logger.info("received request")

        try:

            src, filename = self._create_source(request)

        except Exception as e:
            if self.agent:
                self.agent.server_status.calls_error += 1
            return jsonify({'status': 400, 'message': e})

        logger.info(src)
        stats = self._process_content(src)
        if filename:
            try:
                os.remove(filename)
            except Exception as e:
                logger.warning(f"Cannot remove file {filename}: {e}")

        if self.agent:
            self.agent.server_status.calls_success += 1
        #return jsonify({'status': 200, 'message': 'content uploaded successfully'})
        return jsonify(status=200, message="content uploaded successfully", statistics=stats)

    def _create_source(self, request: request):
        src: Source = None
        filename: str = None
        
        if 'file' in request.files:  # multipart/form-data
            file = request.files['file']
            filename = file.filename
            if filename == "":
                raise ServerException("no file")
            provider = self.provider if self.provider else filename
            ext = (''.join(Path(filename).suffixes))[1:]
            logger.debug(f"{ext=}")
            if ext in Source.registered_extensions:  # extension registred by user
                klass, kwargs = Source.registered_extensions[ext]
                filename = secure_filename(filename)
                file.save(filename)
                src = klass(filename, **kwargs)
            elif ext not in ("txt", "csv", "json"):
                raise ServerException(f"unknown extension {ext}")
            elif ext == 'json':  # JSON extension
                filename = None # here we don't save the file
                data = json.load(file)
                src = SourceJson(data, provider=provider,
                                 jsonpath=self.jsonpath)
            else:  # processed as text
                filename = None # here we don't save the file
                data = file.read().decode('utf-8')
                src = SourceStream(data.splitlines(), provider=provider)
        else:  # raw binary
            if request.is_json:
                logger.info("request is json")
                data = request.get_json()
                src = SourceJson(data, provider=self.provider,
                                 jsonpath=self.jsonpath)
            else:
                logger.info("request is plain text")
                data = request.get_data().decode("utf-8", errors="replace")
                src = SourceStream(data.splitlines())

        return src, filename


class ServerUdp(Server):
    """
    ServerUdp allows receiving UDP frames

    A typical use case is to gather NMEA data from an AIS-receiver.
    """

    def __init__(self,
                 host: str = "127.0.0.1",
                 port: int = 10110,
                 bufsize: int = 1024,
                 provider: str = "UDP Server",
                 ignore_header: bool = False):
        """
        Parameters
        ----------
        host : str
            The server hostname
        port : int
            The server port
        bufsize : int
            Buffer size
        """
        super().__init__(provider, ignore_header)
        self.hostname = host
        self.port = port
        self.bufsize = bufsize
        self.interrupted = False

        logger.info(f"init UDP server addr = {host}:{port}")
        self.s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.s.bind((host, port))
        logger.info(f"UDP server started")
        signal.signal(signal.SIGINT, self.handle_signal)
        signal.signal(signal.SIGQUIT, self.handle_signal)
        signal.signal(signal.SIGTERM, self.handle_signal)

    def run(self):
        logger.info("ready...")
        self.agent.server_status.starttime = datetime.now()
        while not self.interrupted:
            logger.info("receiving...")
            try:
                data, addr = self.s.recvfrom(self.bufsize)
                logger.info(f"received UDP message from {addr} : {data}")
                if self.agent:
                    self.agent.server_status.lastcalltime = datetime.now()
                    self.agent.server_status.calls += 1
                src: Source = SourceSingle(
                    data.decode('utf-8'), provider=self.provider)
                self._process_content(src)
                if self.agent:
                    self.agent.server_status.calls_success += 1
            except Exception as e:
                if not self.interrupted:
                    logger.error(e)
                    if self.agent:
                        self.agent.server_status.calls_error += 1

    def close(self):
        self.s.close()
        logger.info(f"UDP server closed")

    def handle_signal(self, signum, frame):
        """Properly clean resources when a signal is received"""
        logger.info("Received SIGNAL : ")
        logger.info("Stopping loop...")
        self.interrupted = True
        self.s.close()
        time.sleep(1)
